
__author__ = 'Lukas Leufen'
__date__ = '2020-09-21'

import inspect
from typing import Union, Dict

from mlair.helpers import remove_items


class AbstractDataHandler:

    _requirements = []
    _store_attributes = []

    def __init__(self, *args, **kwargs):
        pass

    @classmethod
    def build(cls, *args, **kwargs):
        """Return initialised class."""
        return cls(*args, **kwargs)

    @classmethod
    def requirements(cls):
        """Return requirements and own arguments without duplicates."""
        return list(set(cls._requirements + cls.own_args()))

    @classmethod
    def own_args(cls, *args):
        """Return all arguments (including kwonlyargs)."""
        arg_spec = inspect.getfullargspec(cls)
        list_of_args = arg_spec.args + arg_spec.kwonlyargs
        return remove_items(list_of_args, ["self"] + list(args))

    @classmethod
    def store_attributes(cls) -> list:
        """
        Let MLAir know that some data should be stored in the data store. This is used for calculations on the train
        subset that should be applied to validation and test subset.

        To work properly, add a class variable cls._store_attributes to your data handler. If your custom data handler
        is constructed on different data handlers (e.g. like the DefaultDataHandler), it is required to overwrite the
        get_store_attributs method in addition to return attributes from the corresponding subclasses. This is not
        required, if only attributes from the main class are to be returned.

        Note, that MLAir will store these attributes with the data handler's identification. This depends on the custom
        data handler setting. When loading an attribute from the data handler, it is therefore required to extract the
        right information by using the class identification. In case of the DefaultDataHandler this can be achieved to
        convert all keys of the attribute to string and compare these with the station parameter.
        """
        return list(set(cls._store_attributes))

    def get_store_attributes(self):
        """Returns all attribute names and values that are indicated by the store_attributes method."""
        attr_dict = {}
        for attr in self.store_attributes():
            attr_dict[attr] = self.__getattribute__(attr)
        return attr_dict

    @classmethod
    def transformation(cls, *args, **kwargs):
        return None

    def apply_transformation(self, data, inverse=False, **kwargs):
        """
        This method must return transformed data. The flag inverse can be used to trigger either transformation or its
        inverse method.
        """
        raise NotImplementedError

    def get_X(self, upsampling=False, as_numpy=False):
        raise NotImplementedError

    def get_Y(self, upsampling=False, as_numpy=False):
        raise NotImplementedError

    def get_data(self, upsampling=False, as_numpy=False):
        return self.get_X(upsampling, as_numpy), self.get_Y(upsampling, as_numpy)

    def get_coordinates(self) -> Union[None, Dict]:
        """Return coordinates as dictionary with keys `lon` and `lat`."""
        return None

    def _hash_list(self):
        return []
